<?php
Txp::get('\Textpattern\Tag\Registry')->register('etc_cache');

if (txpinterface == 'admin') {
    register_callback('etc_cache_install', 'plugin_lifecycle.etc_cache');
    new etc_Cache();
} elseif (txpinterface == 'public') {
    global $nolog;
    serverSet('HTTP_USER_AGENT') != 'etc-cache' or $nolog = true;
    if (get_pref('etc_cache_enable')) new etc_Cache();
}

// -------------------------------------------------------------

function etc_cache($atts, $thing = null)
{
    global $prefs, $pretext;
    static $lastmod = null, $times = array(), $filter = null, $reset = null;

    if ($atts === true) return compact('filter', 'reset');

    $oldreset = $reset;

    extract(lAtts(array(
        'id'    => empty($atts['form']) ? null : $atts['form'],
        'url'   => false,
        'form'  => '',
        'reset' => $reset,
        'time'  => true
    ), $atts));

    if ($url === true) {
        $id .= md5($pretext['req']);
        $url = false;
    }

    if (empty($id)) if ($url) {
        $id = md5($url);
    } else {
        $filter = isset($thing) ? parse($thing) : null;

        return;
    }

    $id = doSlash($id);
    $now = time();

    if (!isset($lastmod)) {
        $lastmod = strtotime($prefs['lastmod']);
    }

    if (!get_pref('etc_cache_active', 1)) {
        $update = -1;
    } elseif (empty($time)) {
        $update = 0;
    } elseif ($time === true) {
        $update = $lastmod;
    } elseif (!isset($times[$time = (string)$time])) {
        $update = is_numeric($time) ? $now + (float)$time : (int)strtotime($time, $now);

        if ($update > $now) {
            if (!is_numeric($time)) {
                $date = date_parse($time);

                foreach (array('year', 'month', 'day', 'hour', 'minute', 'second', 'fraction') as $key) {
                    if ($date[$key] !== false) {
                        $update = -1;
                        break;
                    }
                }
            }
        }

        $times[$time] = $update;
    } else {
        $update = $times[$time];
    }

    $cached = $update >= 0 ? safe_row('reset, UNIX_TIMESTAMP(time) AS time, url, text', 'etc_cache', "id = '$id' AND time IS NOT NULL") : false;
    $onlastmod = $cached && ($time === true || $update > $now) && ($cached['reset'] == '%' || $cached['reset'] == '1');

    if ($update > $now) {
        $update = $onlastmod ? max(2*$now - $update, $lastmod) : 2*$now - $update;
//    } elseif ($time !== true) {
//        if (!isset($reset) && !isset($cached['reset'])) $reset = '';
    } elseif ($cached && !$onlastmod && $time === true) {
        $update = 0;
    }

    $parse = !$cached || $update < 0 || $cached['time'] < $update;

    if ($parse) {
        $out = $url ? file_get_contents($url) : false;
        $out !== false or $out = $form ? parse_form($form, $thing) : parse((string)$thing);

        if ($update >= 0) {
            $renew = ($reset || !isset($cached['url']) ? "url='".doSlash($pretext['req'])."'," : '')
                .(isset($reset) && $reset !== true ? "reset='".doSlash($reset)."'," : '')
                .(isset($filter) ? "filter='{".doSlash($filter)."}'," : '');

            safe_upsert('etc_cache', "time = NOW(), $renew text = '".doSlash($out)."'", "id = '$id'");
        }
    } else {
        $out = $cached['text'];
    }

    $reset = $oldreset;

    return $out;
}

/********** Admin class ***********/

class etc_Cache
{
private $files = null, $dir = 'cache';

public function __construct()
{
    if (txpinterface === 'admin') {
      $this->dir = '../'.$this->dir;
      add_privs('etc_cache', '1,2');
      register_tab('extensions', 'etc_cache', gTxt('etc_cache_tab'));
      register_callback(array($this, 'tab'), 'etc_cache');
      if (get_pref('etc_cache_active', 1)) register_callback(array($this, 'update'), 'site.update');
    } elseif (txpinterface === 'public') {
      if (version_compare(get_pref('version'), '4.8.5') < 0) {
        register_callback(array($this, 'dispatch'), 'pretext', '', 1);
      } else {
        register_callback(array($this, 'dispatch'), 'pretext_end', '', 1);
      }
    }
}

public function dispatch() {
  global $production_status;  
  static $run = false;
    if ($run) return;
    $run = true;
    $pretext = pretext(null);
    $req = $pretext['req'];
    if (!empty($_GET) ||
      !empty($_POST) ||
      strpos($req, '/index.php') === 0
    ) return;

    $is_file = rtrim($req, '/') == $req;
    $req = urldecode($req);
        
    if (serverSet('HTTP_USER_AGENT') != 'etc-cache')
      foreach (array('html' => 'text/html') as $ext => $type) {
        $file = $this->dir.'/'.$req.($is_file ? '' : 'index').'.'.$ext;
        
        if (is_readable($file)) {
            set_headers(array('Content-type' => $type));
            die(file_get_contents($file).($production_status === 'debug' ? n.'<!-- served by txp -->' : ''));
        }
      }

      register_callback(array($this, 'cache'), 'textpattern_end');
}

public function cache() {
      global $pretext, $production_status;

      if (!$this->filter()) return;

      $cache = $this->dir;
      $dir = explode('/', ltrim($pretext['req'], '/'));

      foreach ($dir as &$p) if ($p !== '') {
        $p = rawurldecode($p);
        if ($p !== sanitizeForFile($p)) return;
      } else {
        $p = 'index';
      }
      
      unset($p);
      $file = array_pop($dir);
      $dir = $cache.($dir ? '/'.implode('/', $dir) : '');
      if (!is_dir($dir) && !mkdir($dir, 0777, true)) return;

      $finfo = pathinfo($file);
      $extension = isset($finfo['extension']) ? $finfo['extension'] : '';
      $headers = set_headers(array('content-type', true));

      $mimetypes = \Txp::get('Textpattern\Skin\Form')->getMimeTypes() + array('html' => 'text/html');

      if (!empty($headers['content-type']) &&
        (!isset($mimetypes[$extension]) || strpos($headers['content-type'], $mimetypes[$extension]) === false)
      ) {
        foreach($mimetypes as $ext => $type) {
          if (strpos($headers['content-type'], $type) !== false) {
            $file .= $extension === $ext ? '' : '.'.$ext; break;
          }
        }
      }
      $out = ob_get_contents().($production_status === 'debug' ? n.'<!-- static cache -->' : '');

      if (file_put_contents($dir.'/'.$file, $out) !== false) {
        extract(etc_cache(true));

        if (isset($reset) || isset($filter)) {
          safe_upsert('etc_cache',
            "url='".doSlash($pretext['req'])."'"
            .($reset === true ? '' : ", reset='".doSlash(isset($reset) ? $reset : 1)."'")
            .(isset($filter) ? ", filter='{".doSlash($filter)."}'" : ''),
            "id='".doSlash('../'.$dir.'/'.$file)."'"
          );
        }
      }
}

private function filter() {
      global $pretext;
      static $filtered = null;
      
      if (isset($filtered)) return $filtered;
      if ($pretext['status'] != '200') return $filtered = false;

      $filter = parse_ini_string(get_pref('etc_cache_exclude'), true);

      if (!empty($filter['headers'])) {
        foreach((array) set_headers(array_fill_keys(array_keys($filter['headers']), true)) as $name => $value) {
          if (preg_match($filter['headers'][$name], $value)) return $filtered = false;
        }
      }

      if (!empty($filter['pretext'])) {
        foreach($filter['pretext'] as $name => $value) {
          if (isset($pretext[$name]) && preg_match($value, $pretext[$name])) return $filtered = false;
        }
      }

      if (!empty($filter['user'])) {
        if ($userInfo = is_logged_in())
          foreach($filter['user'] as $name => $value) {
            if (isset($userInfo[$name]) && preg_match($value, $userInfo[$name])) return $filtered = false;
          }
      }

      return $filtered = true;
}

private function match($value, array $values) {
    if ($values) foreach((array)$values as $v) {
        if (preg_match('/^'.str_replace('%', '.*', preg_quote($v, '/')).'$/', $value)) return true;
    }

    return false;
}

public function update($event, $step = '', $rs = null)
{
    $urls = $forms = array();
    $safe_step = $step ? doSlash($step) : '';
    $qdir = doSlash($this->dir);
    $filter = "'$safe_step' LIKE reset".($step ? " OR FIND_IN_SET('$safe_step', reset)" : '');
    $files = safe_column(array('id', 'reset'), 'etc_cache', "id LIKE '$qdir%'");

    foreach (safe_rows("id, url, filter", 'etc_cache', "reset>'' AND url>'' AND ($filter)") as $cache) {
        extract($cache);

        if ($step && $filter) {
            $items = json_decode($filter, true);

            if (isset($items[$step]) && is_array($items[$step])) {
                foreach ($items[$step] as $field => $value) {
                    if (isset($rs[$field]) && !$this->match($rs[$field], (array)$value)) {
                        continue 2;
                    }
                }
            }
        }

        $urls[$id] = $url;
        isset($files[$id]) or $forms[] = $id;
        unset($files[$id]);
    }

    foreach ($this->getFiles() as $file => $url) {
      if (!isset($files[$file]) || $files[$file] == '1') {
        unlink($file);
      } else {
        unset($files[$file]);
      }
    }

    if ($urls) {
        $idset = quote_list(array_keys($urls), ',');
        safe_update('etc_cache', 'time = NULL', "id IN ($idset)");
        $this->ping(array_unique($urls));
    }

    $fileset = quote_list(array_keys($files), ',').','.quote_list(array_keys($forms), ',');
    $fileset = trim($fileset, ',');

    if ($fileset) {
        safe_delete('etc_cache', "time IS NULL AND id IN($fileset)");
        safe_optimize('etc_cache');
    }
}

private function getFiles($root = null)
{
//  while ($new = glob($this->dir.'/'.$dir.'*') and $dir .= '*/') {...}

    if (!isset($root)) $root = $this->dir;
    $qdir = preg_quote($this->dir, '/');

    if (is_file($root)) {
        DS == '/' or $root = strtr($root, DS, '/');
        return array($root => preg_replace('/(?:^'.$qdir.'|(?:index)?\.html$)/', '', $root));
    } elseif (!is_dir($root)) {
        return array();
    }

    $files = array_diff(scandir($root), array('.', '..'));
    $result = array();

    foreach ($files as $file) {
        $result += $this->getFiles($root.DS.$file);
    }

    return $result;
}


private function ping ($urls = array()) {
	static $context = null;
	$out = array();
	if (empty($urls)) return $out;
	if (!isset($context)) {
		$context = stream_context_create(array('http'=>array(
			'method'=>'POST',
			'header'=>"Cache-Control: max-age=0, no-store, no-cache\r\n" .
				"Pragma: no-cache\r\n" .
				"User-Agent: etc-cache\r\n"
			)
		));
	}

	foreach ((array)$urls as $url) {
		$uri = preg_match('/^\s*https?:/', $url) ? $url : hu.ltrim($url, '/ ');
		if (file_get_contents($uri, false, $context) !== false) $out[] = $url;
	}

	return $out;
}

private function process($save) {
	extract(psa(array('id', 'url')));
	$qid = quote_list($id);
  $is_file = strpos($id, $this->dir) === 0 && is_file($id);
	$message = gTxt('etc_cache_updated');
	$save or $save = ps('save');

  switch($save) {
    case gTxt('save') : if ($id) safe_upsert('etc_cache',
      "reset='".doSlash(ps('reset'))."',
      url='".doSlash($url)."',
      filter='".doSlash(ps('filter'))."'",
      "id=$qid");
    break;
    case 'Delete' : safe_delete('etc_cache', $id ? "id=$qid" : '1');
    if (!$id) \Txp::get('\Textpattern\Admin\Tools')->removeFiles($this->dir);
    elseif ($is_file) unlink($id);
    break;
    case gTxt('update') : 
    if ($is_file) {
      unlink($id);
    } else {
      safe_update('etc_cache', 'time=NULL', $id ? "id=$qid" : '1');
    }
    if ($id && $url = $is_file ?
        $url :
        safe_field('url', 'etc_cache', "id=$qid")
      ) {
      $this->ping($url);
    }
    break;
    default: $message = '';
  }

  return $message;
}

/*************** admin *****************/

public function tab($event, $step) {
	global $prefs;

	if($step == 'save' && bouncer($step, array('save'=>true))) {
    $message = $this->process(ps('action'));
	} else {
    $message = '';
	}

	$archive_dateformat = $prefs['archive_dateformat'];
  $lastmod = safe_strtotime($prefs['lastmod']);
  $form = 0;

	$rs = safe_column(array('id', 'id, time, url, reset, filter, IF(LENGTH(text)>512, CONCAT(LEFT(text, 504), " ..."), text) AS text'), 'etc_cache', '1 ORDER BY time DESC');

  foreach ($this->getFiles() as $file => $url) {
    $rs[$file] = array(
      'id' => $file,
      'time' => filemtime($file),
      'url' => strtr(urlencode($url), array('%2F' => '/', '%2B' => '+')),
      'reset' => isset($rs[$file]) ? $rs[$file]['reset'] : 1,
      'filter' => isset($rs[$file]) ? $rs[$file]['filter'] : '',
      'text' => $url
    );
  }

  if (get_pref('etc_cache_enable') && !is_dir($this->dir) && !mkdir($dir, 0777, true)) {
      $message = (empty($message) ? '' : $message.br).gTxt('path_creation_or_writing_failed', array('{list}' => $this->dir));
      $message = array($message, 2);
  }

	pagetop(gTxt('etc_cache_tab'), $message);
    
    if ($rs) {
        $contentBlock = n.tag_start('table', array('class' => 'txp-list--no-options')).
            n.tag_start('thead').
            tr(
                n.'<th>'.dLink('etc_cache', 'save', 'save', 'Delete').n.'</th>'.
                n.'<th>ID</th>'.
                n.'<th>'.gTxt('etc_cache_cached_at').'</th>'.
                n.'<th>URL</th>'.
                n.'<th>'.gTxt('etc_cache_reset').'</th>'.
                n.'<th>'.gTxt('etc_cache_filter').'</th>'.
                n.'<th>'.gTxt('etc_cache_actions').'</th>'
            ).
            n.tag_end('thead').
            n.tag_start('tbody');

    	foreach($rs as $row) {
        $form++;
    		extract($row);
    		$is_file = is_int($time);
    		$missing = !isset($time);
    		$time = $is_file ? $time : safe_strtotime($time);
    		$class = (($reset == '%' || $reset == '1') && $time < $lastmod ? ' warning' : '');
    		$date = $missing ? gTxt('status_missing') : doSpecial(safe_strftime($archive_dateformat, $time));

    		$contentBlock .= n.'<tr>'.
          n.'<td>'.n.tag(
              span(gTxt('delete'), array('class' => 'ui-icon ui-icon-close')),
              'button',
              array(
                  'name'      => 'save',
                  'value'      => 'Delete',
                  'class'      => 'destroy',
                  'type'       => 'submit',
                  'title'      => gTxt('delete'),
                  'aria-label' => gTxt('delete'),
                  'form'       => 'form-'.$form
              )
          ).'</td>'.
          n.tag(href(txpspecialchars($id), $is_file ? $id : '..'.$url), 'td', array('title' => $text, 'class' => $missing ? 'warning' : false)).
          n.'<td class="date'.$class.'">'.$date.'</td>'.
          n.'<td>'.fInput('text', array('name' => 'url', 'form' => 'form-'.$form, 'readonly' => $is_file), $url, '', '', '', INPUT_MEDIUM).n.'</td>'.
          n.'<td>'.fInput('text', array('name' => 'reset', 'form' => 'form-'.$form), $reset, '', '', '', INPUT_MEDIUM).n.'</td>'.
          n.'<td>'.fInput('text', array('name' => 'filter', 'form' => 'form-'.$form), $filter, '', '', '', INPUT_MEDIUM).n.'</td>'.
          n.'<td>'.'<form method="post" action="?event=etc_cache" id="form-'.($form).'" name="longform">'.
            fInput('submit', 'save', gTxt('save')).
            fInput('submit', 'save', gTxt('update')).
            eInput('etc_cache').
            sInput('save').
            hInput('id', $id).
            hInput('action', gTxt('save')).
            tInput().
          '</form>'.'</td>'.
        n.'</tr>';
    	}

        $contentBlock .= n.tag_end('tbody').
            n.tag_end('table');
    } else {
        $contentBlock = graf(
            span(null, array('class' => 'ui-icon ui-icon-info')).' '.
            gTxt('etc_cache_no_cached_items'),
            array('class' => 'alert-block information')
        );
    }

        $table = new \Textpattern\Admin\Table();
        echo $table->render(array('heading' => 'etc_cache_heading'), '', '', $contentBlock, '');
        script_js(<<<EOS
            $(document).ready(function ()
            {
                $('#txp-main').on('click', '[name="save"]', function(e) {console.log(e)
                  button = $(this);
                  form = button.attr('form');
                  form = !!form ? $('#'+form) : button.closest('form');
                  form.find('input[name="action"]').val(button.val());
                });
            });
EOS
        , false);

//    $htaccess = file_get_contents(get_pref('path_to_site').'/.htaccess');
}
//------- end class
}

/*************** install *****************/

function etc_cache_install($event='', $step='')
{
	if($step == 'deleted') {
		safe_delete('txp_prefs', "name LIKE 'etc\_cache\_%'");
		safe_query('DROP TABLE IF EXISTS '.safe_pfx('etc_cache'));
		return;
	}
	if($step == 'enabled') {
		$qc="CREATE TABLE IF NOT EXISTS ".safe_pfx('etc_cache')." (";
		$qc.= <<<EOF
			`id` varchar(255) NOT NULL,
			`url` varchar(255) DEFAULT NULL,
			`time` DATETIME DEFAULT NULL,
			`reset` varchar(255) DEFAULT '1',
			`filter` varchar(767) DEFAULT NULL,
			`text` mediumtext,
			PRIMARY KEY (`id`)
			) ENGINE=MyISAM CHARACTER SET=utf8 ;
EOF;
		safe_query($qc);
		if (get_pref('etc_cache_active', null) === null)
      set_pref('etc_cache_active', 1, 'advanced_options', PREF_PLUGIN, 'yesnoradio', 490, PREF_GLOBAL);
		if (get_pref('etc_cache_enable', null) === null)
      set_pref('etc_cache_enable', 0, 'advanced_options', PREF_PLUGIN, 'yesnoradio', 495, PREF_GLOBAL);
		if (get_pref('etc_cache_exclude', false) === false)
      set_pref('etc_cache_exclude', '[headers]
cache-control="/no-store|no-cache|private/i"
[pretext]
#s=excluded_section
[user]
name="/./"', 'advanced_options', PREF_PLUGIN, 'longtext_input', 500, PREF_GLOBAL);
		return;
	}
}
